<?php
/**
 * The IMP_MailboxCache:: class contains all functions related to caching
 * IMAP mailbox information across sessions.
 *
 * $Horde: imp/lib/IMAP/MailboxCache.php,v 1.99.2.4 2008/01/03 07:07:59 slusarz Exp $
 *
 * Copyright 2005-2008 The Horde Project (http://www.horde.org/)
 *
 * See the enclosed file COPYING for license information (GPL). If you
 * did not receive this file, see http://www.fsf.org/copyleft/gpl.html.
 *
 * @author  Michael Slusarz <slusarz@horde.org>
 * @since   IMP 4.2
 * @package IMP
 */
class IMP_MailboxCache {

    /**
     * The Horde_Cache object to use.
     *
     * @var Horde_Cache
     */
    var $_cache;

    /**
     * The list of items to save on shutdown.
     *
     * @var array
     */
    var $_save = array();

    /**
     * The working data for the current pageload.  All changes take place to
     * this data.
     *
     * @var $_data
     */
    var $_data = array();

    /**
     * Attempts to return a reference to a concrete IMP_MailboxCache instance.
     * It will only create a new instance if no IMP_MailboxCache instance with
     * the same parameters currently exists.
     *
     * This method must be invoked as: $var = &IMP_MailboxCache::singleton();
     *
     * @return mixed  The created concrete IMP_MailboxCache instance, or false
     *                on error.
     */
    function &singleton()
    {
        static $instance;

        if (!isset($instance)) {
            $instance = new IMP_MailboxCache();
        }

        return $instance;
    }

    /**
     * Constructor.
     */
    function IMP_MailboxCache()
    {
        if (empty($GLOBALS['conf']['mboxcache']['use_mboxcache'])) {
            return;
        }

        require_once 'Horde/Cache.php';
        require_once 'Horde/Serialize.php';

        /* Initialize the Cache object. */
        $this->_cache = &Horde_Cache::singleton($GLOBALS['conf']['cache']['driver'], Horde::getDriverConfig('cache', $GLOBALS['conf']['cache']['driver']));
        if (is_a($this->_cache, 'PEAR_Error')) {
            Horde::fatal($this->_cache, __FILE__, __LINE__);
        }

        /* Determine the serialization configuration once per session. */
        if (!isset($_SESSION['imp']['mboxcache'])) {
            $_SESSION['imp']['mboxcache'] = array('compress' => null,'prune' => array(), 'lifetime' => empty($GLOBALS['conf']['mboxcache']['lifetime']) ? 0 : $GLOBALS['conf']['mboxcache']['lifetime']);
            if (!empty($GLOBALS['conf']['mboxcache']['use_compress'])) {
                require_once HORDE_BASE . '/lib/version.php';
                switch ($GLOBALS['conf']['mboxcache']['compress']['method']) {
                case 'gzip':
                    if (Horde_Serialize::hasCapability(SERIALIZE_GZ_COMPRESS) &&
                        (version_compare(HORDE_VERSION, '3.1.0') >= 0)) {
                        $_SESSION['imp']['mboxcache']['compress'] = SERIALIZE_GZ_DEFLATE;
                    } else {
                        Horde::logMessage('Could not use gzip compression in IMP_MailboxCache::.', __FILE__, __LINE__, PEAR_LOG_NOTICE);
                    }
                    break;

                case 'lzf':
                    if (defined('SERIALIZE_LZF') &&
                        Horde_Serialize::hasCapability(SERIALIZE_LZF)) {
                        $_SESSION['imp']['mboxcache']['compress'] = SERIALIZE_LZF;
                    } else {
                        Horde::logMessage('Could not use lzf compression in IMP_MailboxCache::.', __FILE__, __LINE__, PEAR_LOG_NOTICE);
                    }
                    break;
                }
            }
        }
    }

    /**
     * Create the unique ID used to store the data in the cache
     *
     * @access private
     *
     * @param string $mailbox  The mailbox to cache.
     *
     * @return string  The cache ID.
     */
    function _getCacheID($mailbox)
    {
        /* Cache ID = username | mailbox */
        return 'imp_mailboxcache_ ' . $_SESSION['imp']['user'] . '|' . $mailbox;
    }

    /**
     * Returns the UID validity timestamp of the given mailbox.
     *
     * @access private
     *
     * @param string $mailbox  The mailbox to query.
     *
     * @return integer  The UID validity timestamp.
     */
    function _getUIDValidity($mailbox)
    {
        require_once IMP_BASE . '/lib/IMAP/Cache.php';
        $imap_cache = &IMP_IMAP_Cache::singleton();
        $status = $imap_cache->getStatus(null, $mailbox);
        return (!empty($status)) ? $status->uidvalidity : false;
    }

    /**
     * Retrieve the imap overview information for the given mailbox and
     * message IDs.
     *
     * @param string $mailbox  An IMAP mailbox string.
     * @param array $uids      The list of message IDS to retrieve overview
     *                         information for.
     * @param integer $mask    A bitmask indicating the fields that should be
     *                         added to the message data. The bitmasks are
     *                         as follows:
     * <pre>
     * 1 =  imap_fetch_overview() information
     *       FIELDS: subject, from, to, date, message_id, references,
     *               in_reply_to, size, uid, flagged, answered, deleted, seen,
     *               draft
     * 2 =   IMAP mailbox arrival information
     *       FIELDS: msgno
     *       NOTE: This option ALWAYS requires an access to the IMAP server to
     *             obtain the message list (sorted by arrival).
     * 4 =   Mailbox information
     *       FIELDS: mailbox
     * 8 =   Mesage structure information (MIME_Message:: object)
     *       FIELDS: structure
     * 16 =  Cached preview data (using prefs value)
     *       FIELDS: preview
     * 32 =  Header information (IMP_Headers:: object)
     *       FIELDS: header
     * 64 =  Cached preview data (overrides prefs value)
     *       FIELDS: preview
     * </pre>
     *
     * @return array  An array of stdClass objects with the UID of the message
     *                as the key; the stdClass objects contain the fields
     *                requested via the $mask parameter or false if the UID
     *                does not exist on the server.
     */
    function retrieve($mailbox, $uids, $mask = 0)
    {
        if (empty($uids) || ($mask == 0)) {
            return array();
        }

        $avail_ids = $this->_loadUIDs($mailbox, $uids);
        $bad_ids = $return_array = $save = array();
        $prev_unread = false;

        $imp_imap = &IMP_IMAP::singleton();
        $imp_imap->changeMbox($mailbox, IMP_IMAP_AUTO);
        $stream = $imp_imap->stream();

        /* Obtain imap_fetch_overview() information. */
        if ($mask & 1) {
            $get_ids = array_diff($uids, $avail_ids);
            foreach ($avail_ids as $val) {
                if (!isset($this->_data[$mailbox][$val]->uid)) {
                    $get_ids[] = $val;
                }
            }

            /* Grab any new overview information we may need. */
            if (!empty($get_ids)) {
                $overview = @imap_fetch_overview($stream, implode(',', $get_ids), FT_UID);
                foreach ($overview as $val) {
                    if (isset($this->_data[$mailbox][$val->uid])) {
                        $ptr = &$this->_data[$mailbox][$val->uid];
                        foreach (get_object_vars($val) as $key => $var) {
                            $ptr->$key = $var;
                        }
                    } else {
                        $this->_data[$mailbox][$val->uid] = $val;
                        $ptr = &$this->_data[$mailbox][$val->uid];
                    }

                    // There should not be any 8bit characters here.  If there
                    // is, we either need to convert from the default charset
                    // or replace with question marks.
                    foreach (array('subject', 'from', 'to') as $val2) {
                        if (!isset($ptr->$val2)) {
                            continue;
                        }

                        if (!empty($GLOBALS['mime_headers']['default_charset'])) {
                            $ptr->$val2 = String::convertCharset($ptr->$val2, $GLOBALS['mime_headers']['default_charset']);
                        } else {
                            $ptr->$val2 = preg_replace('/[\x80-\xff]/', '?', $ptr->$val2);
                        }
                    }

                    if (!empty($this->_cache)) {
                        unset($ptr->msgno, $ptr->recent);
                    }
                    $save[] = $val->uid;
                }

                /* This is the list of IDs that are invalid. */
                if (count($save) != count($get_ids)) {
                    $bad_ids = array_values(array_diff($get_ids, $save));
                    foreach ($bad_ids as $val) {
                        unset($uids[$val]);
                    }
                }
            }
            $prev_unread = $GLOBALS['prefs']->getValue('preview_show_unread');
        }

        /* Add 'structure' information. */
        if ($mask >= 8) {
            foreach ($uids as $val) {
                $ptr = &$this->_data[$mailbox][$val];
                if ($mask & 8) {
                    // Check for invalid data.
                    if (isset($ptr->structure) && !is_object($ptr->structure)) {
                        unset($ptr->structure);
                    }

                    if (!isset($ptr->structure)) {
                        $structure = @imap_fetchstructure($stream, $val, FT_UID);
                        /* Only retrieve if there is a valid structure. */
                        if ($structure) {
                            $ptr->structure = MIME_Structure::parse($structure);
                            $save[] = $val;
                        }
                    }
                }

                if (($mask & 16) || ($mask & 64)) {
                    if (isset($ptr->preview)) {
                        if ($prev_unread && !($mask & 64) && $ptr->seen) {
                            unset($ptr->preview);
                            $save[] = $val;
                        }
                    } else {
                        if (!$prev_unread || !$ptr->seen || ($mask & 64)) {
                            list($ptr->preview, $ptr->preview_cut) = $this->_generatePreview($mailbox, $val);
                            $save[] = $val;
                        }
                    }
                }

                if ($mask & 32) {
                    // Check for invalid data.
                    if (isset($ptr->header) && !is_object($ptr->header)) {
                        unset($ptr->header);
                    }

                    if (!isset($ptr->header)) {
                        $ptr->header = new IMP_Headers($val);
                        $ptr->header->buildHeaders();
                        $save[] = $val;
                    }
                }
            }
        }

        /* Get the return array now. */
        foreach ($uids as $val) {
            if (isset($this->_data[$mailbox][$val])) {
                $return_array[$val] = $this->_data[$mailbox][$val];
                if ($mask & 4) {
                    $return_array[$val]->mailbox = $mailbox;
                }
            }
        }

        /* Determine if we need to add additional fields to the data. */
        if (!empty($this->_cache) && ($mask & 2)) {
            require_once IMP_BASE . '/lib/IMAP/Cache.php';
            $imap_cache = &IMP_IMAP_Cache::singleton();
            foreach ($imap_cache->getMailboxArrival($mailbox) as $key => $val) {
                if (isset($return_array[$val])) {
                    $return_array[$val]->msgno = ++$key;
                }
            }
        }

        if (!empty($save)) {
            $this->_saveMailbox($mailbox, array_unique($save));
        }

        /* Readd "bad" IDs now. */
        foreach ($bad_ids as $val) {
            $return_array[$val] = false;
        }

        return $return_array;
    }

    /**
     * Load the given mailbox either by regenerating from the cache or by
     * using the current in-memory cache.
     *
     * @access private
     *
     * @param string $mailbox  The mailbox to load.
     *
     * @return boolean  True on success, false on failure.
     */
    function _loadMailbox($mailbox)
    {
        /* Need these to avoid unloaded library errors when unserializing. */
        require_once 'Horde/MIME/Structure.php';
        require_once IMP_BASE . '/lib/MIME/Headers.php';

        /* Check the _save array first. */
        if (!empty($this->_data[$mailbox])) {
            return true;
        } elseif (!empty($this->_cache)) {
            /* Get the unique cache identifier for this mailbox. */
            $cache_id = $this->_getCacheID($mailbox);

            /* Attempt to grab data from the cache. */
            $data = $this->_cache->get($cache_id, $_SESSION['imp']['mboxcache']['lifetime']);
            if ($data != false) {
                $this->_data[$mailbox] = Horde_Serialize::unserialize($data, SERIALIZE_BASIC);
                if (!is_array($this->_data[$mailbox])) {
                    $this->_data[$mailbox] = array();
                    return false;
                }

                /* Check UID validity and do garbage collection. */
                $uidvalid = $this->_getUIDValidity($mailbox);
                if (!empty($this->_data[$mailbox]['__uidvalid']) &&
                    ($this->_data[$mailbox]['__uidvalid'] != $uidvalid)) {
                    $this->_cache->expire($cache_id);
                    unset($this->_data[$mailbox]);
                } else {
                    /* Perform garbage collection once a login. */
                    if (!isset($_SESSION['imp']['mboxcache']['prune'][$mailbox])) {
                        $imap_cache = &IMP_IMAP_Cache::singleton();
                        $all_uids = $imap_cache->getMailboxArrival($mailbox, false);
                        foreach (array_diff(array_keys($this->_data[$mailbox]), $all_uids) as $val) {
                            if (is_integer($val)) {
                                unset($this->_data[$mailbox][$val]);
                            }
                        }
                        $this->_saveMailbox($mailbox);
                        $_SESSION['imp']['mboxcache']['prune'][$mailbox] = true;
                    }
                    return true;
                }
            }
        }

        $this->_data[$mailbox] = array();
        return false;
    }

    /**
     * Given a list of UIDs, unpacks the messages from stored cache data and
     * returns the list of UIDs that exist in the cache.
     *
     * @access private
     *
     * @param string $mailbox  The current mailbox.
     * @param array $uids      The list of UIDs to load.
     *
     * @return array  The list of UIDs with cache entries.
     */
    function _loadUIDs($mailbox, $uids)
    {
        if (!$this->_loadMailbox($mailbox)) {
            return array();
        }

        $loaded = array();
        $ptr = &$this->_data[$mailbox];

        foreach ($uids as $val) {
            if (isset($ptr[$val])) {
                if (is_object($ptr[$val])) {
                    $loaded[] = $val;
                } elseif (!is_null($_SESSION['imp']['mboxcache']['compress'])) {
                    $ptr[$val] = Horde_Serialize::unserialize($ptr[$val], array($_SESSION['imp']['mboxcache']['compress'], SERIALIZE_BASIC));
                    if (is_a($ptr[$val], 'PEAR_Error')) {
                        unset($ptr[$val]);
                    } else {
                        $loaded[] = $val;
                    }
                } else {
                    unset($ptr[$val]);
                }
            }
        }

        return $loaded;
    }

    /**
     * Register mailbox data to be saved to cache at the end of a pageload.
     *
     * @access private
     *
     * @param string $mailbox  The mailbox to save.
     * @param array $uids      A specific list of UIDs that have been altered.
     *                         in the given mailbox.
     */
    function _saveMailbox($mailbox, $uids = array())
    {
        if (!empty($this->_cache)) {
            if (empty($this->_save)) {
                register_shutdown_function(array(&$this, '_addCacheShutdown'));
            }
            if (!isset($this->_data[$mailbox]['__uidvalid'])) {
                $this->_data[$mailbox]['__uidvalid'] = $this->_getUIDValidity($mailbox);
            }
            if (!isset($this->_save[$mailbox])) {
                $this->_save[$mailbox] = array();
            }
            $this->_save[$mailbox] = array_merge($this->_save[$mailbox], $uids);
        }
    }

    /**
     * Update message flags in the cache.
     *
     * @param string $mailbox  The mailbox.
     * @param array $uids      The list of message UIDs to update.
     * @param array $flags     The flags to set. Valid arguments: seen,
     *                         deleted, flagged, answered, draft
     * @param boolean $set     True to set the flag, false to clear flag.
     */
    function updateFlags($mailbox, $uids, $flags, $set)
    {
        $uids = $this->_loadUIDs($mailbox, $uids);
        if (empty($uids)) {
            return;
        }

        foreach ($uids as $id) {
            foreach (array_map('strtolower', $flags) as $val) {
                $this->_data[$mailbox][$id]->$val = intval($set);
            }
        }

        $this->_saveMailbox($mailbox, $uids);
    }

    /**
     * Delete messages in the cache.
     *
     * @param string $mailbox  The mailbox.
     * @param array $uids      The list of message UIDs to delete.
     */
    function deleteMsgs($mailbox, $uids)
    {
        if (!$this->_loadMailbox($mailbox)) {
            return;
        }

        $save = empty($this->_cache);

        foreach ($uids as $id) {
            if (!$save || isset($this->_data[$mailbox][$id])) {
                $save = true;
            }
            unset($this->_data[$mailbox][$id]);
        }

        if ($save) {
            $this->_saveMailbox($mailbox);
        }
    }

    /**
     * Saves items to the cache at shutdown.
     *
     * @access private
     */
    function _addCacheShutdown()
    {
        foreach ($this->_save as $key => $val) {
            $ptr = &$this->_data[$key];
            $cacheid = $this->_getCacheID($key);

            /* Don't save empty entries (there will always be 1 entry for
             * the uidvalid entry). */
            if (count($ptr) == 1) {
                $this->_cache->expire($cacheid);
                continue;
            }

            /* Compress individual UID entries if they have changed. We will
               worry about error checking when decompressing (cache data will
               automatically be invalidated then). */
            if ($_SESSION['imp']['mboxcache']['compress'] && !empty($val)) {
                foreach (array_unique($val) as $uid) {
                    if (isset($ptr[$uid])) {
                        $ptr[$uid] = Horde_Serialize::serialize($ptr[$uid], array(SERIALIZE_BASIC, $_SESSION['imp']['mboxcache']['compress']));
                    }
                }
            }

            $this->_cache->set($cacheid, Horde_Serialize::serialize($ptr, SERIALIZE_BASIC), $_SESSION['imp']['mboxcache']['lifetime']);
        }
    }

    /**
     * Generate the preview text.
     *
     * @access private
     *
     * @param string $mailbox  The mailbox.
     * @param integer $uid     The UID to generate a preview for.
     *
     * @return array  Array of the preview text and a flag if the preview has
     *                been cut down.
     */
    function _generatePreview($mailbox, $uid)
    {
        $ptext = '';
        $cut = false;

        $imp_imap = &IMP_IMAP::singleton();
        $imp_imap->changeMbox($mailbox, IMP_IMAP_READONLY);

        require_once IMP_BASE . '/lib/MIME/Contents.php';
        $imp_contents = &IMP_Contents::singleton($uid . IMP_IDX_SEP . $mailbox);
        if (is_a($imp_contents, 'PEAR_Error')) {
            return array('', false);
        }

        if (($mimeid = $imp_contents->findBody()) !== null) {
            $pmime = $imp_contents->getDecodedMIMEPart($mimeid);
            $ptext = $pmime->getContents();
            $ptext = String::convertCharset($ptext, $pmime->getCharset());
            if ($pmime->getType() == 'text/html') {
                require_once 'Horde/Text/Filter.php';
                $ptext = Text_Filter::filter($ptext, 'html2text');
            }

            $maxlen = (empty($GLOBALS['conf']['mboxcache']['preview_size'])) ? $GLOBALS['prefs']->getValue('preview_maxlen') : $GLOBALS['conf']['mboxcache']['preview_size'];
            if (String::length($ptext) > $maxlen) {
                $ptext = String::substr($ptext, 0, $maxlen) . ' ...';
                $cut = true;
            }
        }

        return array($ptext, $cut);
    }

    /**
     * Deletes a mailbox from the cache.
     *
     * @param array $mboxes  The list of mailboxes to delete.
     */
    function deleteMboxes($mboxes)
    {
        foreach ($mboxes as $val) {
            if (!empty($this->_cache)) {
                $this->_cache->expire($this->_getCacheID($val));
            }
            unset($this->_data[$val]);
        }
    }

}
